/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "imaging_data.h"
#include "oirSupport.h"
#define clight  2.9979250e8 /*m per sec */



/*
   data structure to handle multiple input files in raw formats
*/
typedef struct 
{
   char **fileNames;
   int  iFile, nFiles;
   long iRow, nRows;
   fitsfile *inFile;
   imaging_data_table_def  *rdata;
   imaging_fdata_table_def *fdata;
   imaging_detector_table_def *detector;
   imaging_data_row   *rData;
   int nregion;
   int rawInput;
   int iChan1, iChan2;
   int nx, ny;
   long  npix;
} inFileInfo;

void Usage() {
    printf("\noirFormFringes infile outfile [-smooth nSmooth] [-averageOrder order] [-removeAverage]\n");
    printf("   infile  = name of file with output from oir1dCompressData \n");
    printf("   outfile = name of output file\n");
    printf("   nSmooth = width (frames) of high-pass filter region (default 50)\n");
    printf("     If nSmooth <=0 no high-pass filter is applied\n");
    printf("   order [if specified&>=0]: remove polynomial of this order as fn of wavenumber from each spectrum \n");
    printf("   removeAverage; backward compatible alternative to order.\n"); 
    printf("      if specified and no order specified then order = 0\n");
    printf("      default: no spectral filtering\n");
    return;
}

/******************************************************************
 *   copy data from a raw (short int) data row to a float data row
 ******************************************************************/
void raw2Float(imaging_data_row *rawData, 
   imaging_fdata_row *inData, int iChan1, int iChan2, int nx, int ny) {

   int ix, iy;
   for (iy=0;iy<ny;iy++) for (ix=0;ix<nx;ix++) {
      inData->data[iChan1][ix+nx*iy] = (float)rawData->data[iChan1][ix+nx*iy]; 
      inData->data[iChan2][ix+nx*iy] = (float)rawData->data[iChan2][ix+nx*iy]; 
      }
/*
            copy nonData info
*/
   inData->time = rawData->time;
   inData->opd[0] = rawData->opd[0];
   inData->opd[1] = rawData->opd[1];
   inData->local_opd[1] = rawData->local_opd[1];
   inData->local_opd[0] = rawData->local_opd[0];
   inData->exp          = rawData->exp;
   inData->offset[0]    = rawData->offset[0];
   inData->offset[1]    = rawData->offset[1];
   inData->rotation     = rawData->rotation;
   inData->frame        = rawData->frame;
   inData->step         = rawData->step;
   inData->reference    = rawData->reference;
   inData->targtype[iChan1]  = rawData->targtype[iChan1];  
   inData->targtype[iChan2]  = rawData->targtype[iChan2];  
}

 void allocateSmoothBuffers(char *fileName, int nSmooth, imaging_fdata_row **buff, 
   int *ierr) {

   fitsfile *inFile ;
   imaging_fdata_table_def *idata;

   int i, iSmooth;
   int ncol, icol;
   char* DATA = "DATA";
/*                  open input file in order to get information
 *                  on fdata table structure
*/
    i = fits_open_file(&inFile, fileName, READONLY, ierr);
    if (*ierr != 0) {
       printf("ERROR:fits openfile of %s gave error %i",fileName, *ierr); 
       printf("%s \n",strerror(errno));
       return ;
    }
    idata= open_imaging_fdata_table(inFile, ierr); 
    i=fits_close_file(inFile, ierr); 
/*                 find data definitions and force to be float
*/
    ncol = idata->def->ncol; 
    for (i=0;i<ncol;i++) 
       if (!strncmp(idata->def->coltype[i], DATA, 4)) idata->def->coldatatype[i] = TFLOAT;
/*                 now create the table structures from the definition */
    *ierr = 0;
    for (iSmooth=0;iSmooth<nSmooth;iSmooth++) buff[iSmooth] = 
       make_imaging_fdata_row(idata);
/*                 throw away definition       */
    kill_imaging_fdata_table_def(idata);
} /* end allocate smooth buffers */

int  readInputRow(inFileInfo *fileInfo, imaging_fdata_row *outData, int *ierr){

   int i;
   *ierr = 0;

   if (fileInfo->iRow >= fileInfo->nRows)  { /* end of file */
      if (fileInfo->iFile >=0) {
         if(fileInfo->rawInput) i = close_imaging_data_table( fileInfo->inFile, fileInfo->rdata, ierr);
         else i = close_imaging_fdata_table( fileInfo->inFile, fileInfo->fdata, ierr);
      }
      if (++fileInfo->iFile >= fileInfo->nFiles)  {
         if (fileInfo->rawInput) kill_imaging_data_row(fileInfo->rData);
         return 1; 
      }/* all done */
/*
         open next input file
*/
      fileInfo->iRow = 0;
      fits_open_file(&fileInfo->inFile, fileInfo->fileNames[fileInfo->iFile], 
         READONLY, ierr);
      if (*ierr != 0) {
         printf("ERROR:fits openfile of %s gave error %i",
            fileInfo->fileNames[fileInfo->iFile], *ierr);
         return 2;
      }
/*        open new input table   */
      if (fileInfo->rawInput) {
         fileInfo->rdata = open_imaging_data_table(fileInfo->inFile, ierr);
         for (i=0;i<fileInfo->nregion;i++) fits_set_tscale(fileInfo->inFile, 
            1+fileInfo->rdata->data_col[i], 1.0, 0.0, ierr);
         if (fileInfo->iFile == 0) fileInfo->rData = make_imaging_data_row(fileInfo->rdata);
      } else {
         fileInfo->fdata = open_imaging_fdata_table(fileInfo->inFile, ierr);
      }
      i = fits_get_num_rows(fileInfo->inFile, &fileInfo->nRows, ierr);
      fprintf(stderr,"   Reading %ld rows from file %i \r", 
         fileInfo->nRows, 1+fileInfo->iFile);
   }   /* ready for new file */
/*
       read the next row; if raw copy to output row and convert to float
*/
      ++fileInfo->iRow;
      if (fileInfo->rawInput) {
         i = read_imaging_data_table(fileInfo->inFile, fileInfo->rdata, fileInfo->rData, 
            fileInfo->iRow, ierr);
         if (*ierr != 0) {
            printf("ERROR: reading data table row %li error %i",fileInfo->iRow, *ierr);
            return 2;
         } 

         raw2Float(fileInfo->rData, outData, fileInfo->iChan1, fileInfo->iChan2, 
            fileInfo->nx, fileInfo->ny); 

      } else {
         i = read_imaging_fdata_table(fileInfo->inFile, fileInfo->fdata, outData, 
            fileInfo->iRow, ierr);
         if (*ierr != 0) {
         printf("ERROR: reading data table row %li error %i",fileInfo->iRow, *ierr);
            return 2;
         }
      }
   return 0;
} /* end of read input row */

void removePoly(imaging_fdata_row *inData, int nx, int ny, int nWin, 
   float* waveNo, int order, float *weight, float *polyCoeff, float *polyFit) {

   int ix, iy, ii;
   int iWin;

   for (iWin=0;iWin<nWin;iWin++) {
      for (iy=0; iy<ny;iy++) {
         ii   = iy*nx;
         minrtsPolyFit(order, nx, waveNo, inData->data[iWin]+ii, weight, polyCoeff);
         minrtsPolyExpand(order, nx, waveNo, polyCoeff, polyFit);
         for (ix=0;ix<nx;ix++) 
            if (weight[ix] > 0) inData->data[iWin][ii+ix] = inData->data[iWin][ii+ix] - polyFit[ix];
     /*       else inData->data[iWin][ii+ix] = 0; */
      }
   }
   return;

}  /* end of removePoly */


void outputRowLessAve(int subtractSmooth, imaging_fdata_row *inData, 
   imaging_fdata_row *outData, long nx, float dit, float** average, int nWin) {

   long ix;
   float xdit;
   int iWin;

   if (dit > 0.) xdit = 1./dit; else xdit = 1.;
/*
            copy nonData info
*/
   outData->time   = inData->time;
   outData->opd[0] = inData->opd[0];
   outData->opd[1] = inData->opd[1];
   outData->local_opd[1] = inData->local_opd[1];
   outData->local_opd[0] = inData->local_opd[0];
   outData->exp = inData->exp;
   outData->offset[0] = inData->offset[0];
   outData->offset[1] = inData->offset[1];
   outData->rotation  = inData->rotation;
   outData->frame     = inData->frame;
   outData->step      = inData->step;
   outData->reference = inData->reference;
   outData->targtype[0] = inData->targtype[0];  
/*
            copy data and remove average multiply by exposure time
*/
   if (subtractSmooth) {
      for (iWin=0;iWin<nWin;iWin++)
         for (ix=0;ix<nx;ix++) outData->data[iWin][ix] =
           (inData->data[iWin][ix] - average[iWin][ix])*xdit;
   } else  {
      for (iWin=0;iWin<nWin;iWin++)
         for (ix=0;ix<nx;ix++) outData->data[iWin][ix] = 
            inData->data[iWin][ix] *xdit;
   }
   return;
} /* end of outputRowLessAve */

int main (int argc, char** argv) {

   int nregion, iregion, outRegion;
   int i, j, ierr;
   long inRows, oRow;
   int nx, ny ;
   long npix;
   float *totalBuffer[2], *averageBuffer[2], *saveOldBuffer[2];

   int nSmooth;
   int iSmooth;
   int iChan1, iChan2;
   int nextIn, nextOut;
   int iKey, maxstep;
   int iarg;
   int subtractAverage, subtractOrder;
   float *polyCoeff, *polyFit;
   float *waveNo, *weight;
   int subtractSmooth;
   int done;

   char outFileName[130];
   char history[80];
   int  iHistory;
   char *bang = "!";

   inFileInfo fileInfo;


   fitsfile *inFile  = NULL;
   fitsfile *outFile = NULL;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *outDetectorDef;
   imaging_fdata_row **inData;
   imaging_fdata_row *outData;

   float dit;

   ierr = 0;
/*    
      set defaults and read optional args
*/
   iarg            = 3;
   subtractAverage = 0;
   subtractSmooth  = 1;
   nSmooth         = 50;
   subtractOrder   = -1;

   if (argc < 3) {
      Usage();
      return 0;
   }
   printf("      ****** oirFormFringes ******\n");
   while (iarg < argc ) {
      if (!strcmp("-smooth", argv[iarg])) sscanf(argv[++iarg], "%i", &nSmooth);
      if (!strcmp("-removeAverage", argv[iarg])) subtractAverage = 1;
      if (!strcmp("-averageOrder", argv[iarg]))  {
         sscanf(argv[++iarg], "%i", &subtractOrder);
         subtractAverage = 1;
      }
      ++iarg;
   }
   if (nSmooth <= 0) {
      subtractSmooth = 0;
      printf("   NOT applying high-pass filter.\n");
      nSmooth        = 1; /* so rest of program doesn't crash */
   } else 
      printf("   width of high-pass window    = %i frames\n", nSmooth);
   if (subtractOrder >3) {
      printf("   I dont accept polynomial orders >3, you specified %i.\n", subtractOrder);
      return 1;
   }
   if (subtractAverage) {
      if (subtractOrder <0) subtractOrder = 0; /* default */
      printf("   Subtracting polynomial spectral fit of order %i.\n",subtractOrder); }
   else printf("   NOT subtracting polynomial fit to spectra.\n");
/*
      get  input file list
*/
   fileInfo.fileNames = oirStrSplit(argv[1], &fileInfo.nFiles);
   printf("   Reading input data from %i file(s):\n",fileInfo.nFiles);
   for (i=0; i<fileInfo.nFiles; i++) {printf("      %s \n", fileInfo.fileNames[i]);}
   ierr = 0;
/*
      create space for smoothing buffers
*/
   inData = (imaging_fdata_row**)malloc(nSmooth*sizeof(imaging_fdata_row *));
   allocateSmoothBuffers(fileInfo.fileNames[0], nSmooth, inData, &ierr); 
   fits_open_file (&inFile, fileInfo.fileNames[0], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i",fileInfo.fileNames[0], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[2]);
   fits_create_file (&outFile, outFileName, &ierr);

   if (ierr !=0) {
      printf ("ERROR: opening fits output %s %i\n",argv[2], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
      copy primary header from input to output
      add some history records
*/
   i = fits_copy_header(inFile, outFile, &ierr);
   snprintf(history,80,"oirFormFringes %i INPUT files", fileInfo.nFiles);
   i = fits_write_history(outFile, history, &ierr); 
   for (i=0; i<fileInfo.nFiles; i++) {
      iHistory = strlen(fileInfo.fileNames[i]);
      iHistory = iHistory - 72;
      if (iHistory < 0) iHistory = 0;
      snprintf(history,80,"%s", fileInfo.fileNames[i]+iHistory);
      j = fits_write_history(outFile, history, &ierr); 
   }

   if (subtractSmooth) snprintf(history,80,"oirFormFringes NSMOOTH= %i", nSmooth);
      else snprintf(history,80,"oirFormFringes NO highpass filter");
   i = fits_write_history(outFile, history, &ierr);
   if (subtractOrder >= 0)
      snprintf(history,80,"oirFormFringes PolyNomial subtract order %i",subtractOrder);
   else
      snprintf(history,80,"oirFormFringes NOT SUBTRACTING Polynomial");
   i = fits_write_history(outFile, history, &ierr);
/*
      get useful keywords from input primary header
*/
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &dit, 0, &ierr);
   ierr = 0;

/*
      get detector info then close input
*/
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr != 0) {
      printf("ERROR: reading detector table %i",ierr);
      return 1;
   }
   nx      = outDetectorDef->naxis[0][0];
   ny      = outDetectorDef->naxis[0][1];
   nregion = outDetectorDef->nregion;
   fileInfo.nregion  = nregion;
   fileInfo.rawInput = (ny > 1);
   if (fileInfo.rawInput) printf("   input data is                  RAW\n");
   else   printf("   input data is                  COMPRESSED\n"); 

   i = fits_close_file(inFile, &ierr);
   printf("   Writing output to %s\n", argv[2]);
/*            decide which channels are interferometric  */
   if (nregion == 2) {
      iChan1 = 0;
      iChan2 = 1;
   } else {
      iChan1 = 1;
      iChan2 = 2;
   }
   fileInfo.iChan1 = iChan1;
   fileInfo.iChan2 = iChan2;

   if (fileInfo.rawInput) outRegion = 2; else outRegion = 1;
   outDetectorDef->nregion = outRegion;
   npix = nx * (long)ny;
   fileInfo.npix = npix;
   fileInfo.nx   = nx;
   fileInfo.ny   = ny;
   iKey = table_find_key(outDetectorDef->def, "NREGION");
   table_put_keyword (outDetectorDef->def, iKey, (void *)&outRegion, &ierr);
   odata = make_imaging_fdata_table_def (outFile, outDetectorDef, 40, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: creating data definition\n");
      return 1;
      }
   outData = make_imaging_fdata_row(odata);

/*
         create smoothing buffers
*/ for (iregion=0;iregion<outRegion;iregion++) {
      totalBuffer[iregion]   = (float *)malloc(npix * sizeof(float));
      averageBuffer[iregion] = (float *)malloc(npix * sizeof(float));
      saveOldBuffer[iregion] = (float *)malloc(npix * sizeof(float));
      memset(totalBuffer[iregion],   0, npix*sizeof(float));
      memset(averageBuffer[iregion], 0, npix*sizeof(float));
      memset(saveOldBuffer[iregion], 0, npix*sizeof(float));
   }
/*
 *       create areas for polynomial data if required
 */
   if (subtractAverage) {
      polyCoeff = (float*)malloc((1+subtractOrder)*sizeof(float));
      polyFit   = (float*)malloc(nx*sizeof(float));
      weight    = (float*)malloc(nx*sizeof(float));
      waveNo    = imaging_detector_wcs(outDetectorDef, 1);
/*
         It actually turns out that applying a weight limited to 8.0 mu < lambda < 13.0 mu
         can lead to changes in the spectral slope. Therefore use the entire wavelength
         range, as in older EWS versions.
*/
/*      for (i=0;i<nx;i++) {
         waveNo[i] = waveNo[i]/clight;
         if ((waveNo[i] > (1./8.e-6)) || (waveNo[i] < (1./13.5e-6))) 
            weight[i] = 0.; else weight[i] = 1.; 
      } */
      for (i=0;i<nx;i++) weight[i] = 1.;
   }
/*
         initial fill of read buffers
         read a row, difference interferometric channels, accumulate->totalBuffer
*/
   iSmooth        =    0;
   fileInfo.iFile =   -1;
   fileInfo.iRow  =    0;
   fileInfo.nRows = -100;
   inRows         =    0;

   while(iSmooth < nSmooth) {
      done = readInputRow(&fileInfo, inData[iSmooth], &ierr);
      if (done == 2) {
         printf("ERROR: reading data table row %li error %i",fileInfo.iRow, ierr);
         return 1;
      }
      if (inData[iSmooth]->targtype[iChan1] == 'T') {
         if (!fileInfo.rawInput) {  /* compressed data-> subtract chan2 from chan 1 */
            for (j=0; j<npix; j++) 
               inData[iSmooth]->data[0][j] = 
               inData[iSmooth]->data[iChan1][j] - 
                  inData[iSmooth]->data[iChan2][j];
             
         } else {  /* copy interferometric data into data[0] and data[1] */
            if (iChan1!=0) memcpy(inData[iSmooth]->data[0], 
               inData[iSmooth]->data[iChan1], npix*sizeof(float));
            if (iChan2!=1) memcpy(inData[iSmooth]->data[1], 
               inData[iSmooth]->data[iChan2], npix*sizeof(float));
         }  /* raw input */
/*              accumulate data rows in totalBuffer */
         for (iregion=0;iregion<outRegion;iregion++) for (j=0;j<npix;j++) 
            totalBuffer[iregion][j]  = 
               totalBuffer[iregion][j] + inData[iSmooth]->data[iregion][j];
         iSmooth = iSmooth+1;
      } /* valid data */
      inRows++;
   } /* iSmooth loop */
/*                convert total to average */
   for (iregion=0;iregion<outRegion;iregion++) for (j=0; j<npix; j++) 
      averageBuffer[iregion][j] = totalBuffer[iregion][j]/nSmooth;
/*
         output nSmooth/2 rows with subtraction of average
*/
   oRow = 1;
   for (nextOut =0; nextOut<(nSmooth/2); nextOut++) {
      outputRowLessAve(subtractSmooth, inData[nextOut], outData, npix, 
         dit, averageBuffer, outRegion);
      if (subtractAverage) removePoly(outData, nx, ny, outRegion, waveNo, subtractOrder, weight,
         polyCoeff, polyFit); 
      i = write_imaging_fdata_table(outFile, odata, outData, 
         oRow++, &ierr);
      if (ierr != 0) {
         printf("ERROR: writing data table row %li error %i",oRow, ierr);
         return 1;
      }
    }
/*       
         now loop over rest of input data, replacing old row with
         new row in data buffer and in accumulator until done
*/
   nextIn = 0;
   for (iregion=0;iregion<outRegion;iregion++)
      memcpy(saveOldBuffer[iregion], inData[nextIn]->data[iregion], 
         npix*sizeof(float));
   while (done == 0) {
/*
        read new row, difference channels and accumulate
*/
      done = readInputRow(&fileInfo, inData[nextIn], &ierr);
      if (done == 2) {
         printf("ERROR: reading data table row %li error %i",fileInfo.iRow, ierr);
         return 1;
      }
      if (!done) {
/*
          check if correct chopping phase
*/
         if (inData[nextIn]->targtype[iChan1] == 'T') {
            if (!fileInfo.rawInput) {
               for (j=0; j<npix; j++) 
                  inData[nextIn]->data[0][j] = 
                     inData[nextIn]->data[iChan1][j] - 
                     inData[nextIn]->data[iChan2][j];
               
            } else {
               if (iChan1 != 0) memcpy(inData[nextIn]->data[0], 
                  inData[nextIn]->data[iChan1], npix*sizeof(float));
               if (iChan2 != 1) memcpy(inData[nextIn]->data[1], 
                  inData[nextIn]->data[iChan2], npix*sizeof(float));
            }
            inRows++;
/*               in totalBuffer add in the row we just read but
 *               subtract the one we just discarded (into saveOldBuffer) */

            for (iregion=0;iregion<outRegion;iregion++) for (j=0; j<npix; j++)
               totalBuffer[iregion][j] = totalBuffer[iregion][j] + 
                  inData[nextIn]->data[iregion][j] - saveOldBuffer[iregion][j];
               
/*
        totalbuffer/nSmooth ->average buffer, output row 
        nextOut after subtraction of average, and update counters 
*/
            for (iregion=0;iregion<outRegion;iregion++) for (j=0; j<npix; j++) 
               averageBuffer[iregion][j] = totalBuffer[iregion][j]/nSmooth;
            outputRowLessAve(subtractSmooth, inData[nextOut], outData, npix, 
               dit, averageBuffer, outRegion);
            if (subtractAverage) removePoly(outData, nx, ny, outRegion, waveNo, subtractOrder, weight,
               polyCoeff, polyFit); 
            i = write_imaging_fdata_table(outFile, odata, outData, 
               oRow++, &ierr);
            if (ierr != 0) {
               printf("ERROR: writing data table row %li error %i",oRow, ierr);
               return 1;
            }
            if (++nextIn == nSmooth)  nextIn  = 0;
            for (iregion=0;iregion<outRegion;iregion++)
               memcpy(saveOldBuffer[iregion], inData[nextIn]->data[iregion], 
                  npix*sizeof(float));
            if (++nextOut == nSmooth) nextOut = 0;
         }
      }   /*  read legitimate row, not end of file */
   }  /*  loop over remaining input rows */
/*
            finish output rows
*/
   while (oRow <=inRows) {
      outputRowLessAve(subtractSmooth, inData[nextOut++], outData, npix, 
         dit, averageBuffer, outRegion);
      if (subtractAverage) removePoly(outData, nx, ny, outRegion, waveNo, subtractOrder, weight,
         polyCoeff, polyFit); 
      i = write_imaging_fdata_table(outFile, odata, outData, 
         oRow++, &ierr);
      if (ierr != 0) {
         printf("ERROR: writing data table row %li error %i",oRow, ierr);
         return 1;
      }
      if (nextOut >= nSmooth) nextOut = 0;
   }
   printf("   number of processed frames   = %ld\n",oRow-1);
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(outData);

   for (i=0;i<nSmooth;i++) kill_imaging_fdata_row(inData[i]); 
   free (inData);
   for (iregion=0;iregion<outRegion;iregion++){
      free (totalBuffer[iregion]);
      free (averageBuffer[iregion]);
      free (saveOldBuffer[iregion]);
   }
   if (subtractAverage) {
      free(polyCoeff);
      free(polyFit);
      free(waveNo);
      free(weight);
   }
   return 0;
}
